
#include <assert.h>

#include "main.h"

#define REPS 4096
#define WIN_WIDTH 320
#define WIN_HEIGHT 320
#undef EXTENDED_TESTS
#define DO_SDL 1

typedef struct _xrender_surf {
  int w, h;
  int depth;
  Visual *vis;
  Drawable draw;
  Picture pic;
  int allocated:1;
} Xrender_Surf;

static struct {
  Xrender_Surf *win, *off, *img;
} xrender;

static struct {
  Imlib_Image win, img;
} imlib;

static struct {
  SDL_Surface *img, *win, *display;
} sdl;

static struct {
  GdkPixbuf *win, *img;
  GtkWidget *da;
} pixbuf;

static struct {
  struct osda *win;
  struct Sprite *img;
} ftwl;

static Display *disp = NULL;
static Window win;

static double get_time(void);
static void time_test(char *description, void (*func) (void));

static Xrender_Surf *xrender_surf_new(Display * disp, Drawable draw, Visual * vis,
			       int w, int h, int alpha);
static Xrender_Surf *xrender_surf_adopt(Display * disp, Drawable draw, Visual * vis,
				 int w, int h);
//static void xrender_surf_free(Display * disp, Xrender_Surf * rs);
static void xrender_surf_populate(Display * disp, Xrender_Surf * rs, int w, int h,
			   int *img_data);
static void xrender_surf_blend(Display * disp, Xrender_Surf * src,
			Xrender_Surf * dst, int x, int y, int w, int h,
			int smooth);

static void populate_from_file(Display * disp, Xrender_Surf * rs,
			       const char *file);
static void main_loop(const char *transparent_filename);
static void setup_window(void);

static double get_time(void)
{
  struct timeval timev;

  gettimeofday(&timev, NULL);
  return (double) timev.tv_sec + (((double) timev.tv_usec) / 1000000);
}

static void time_test(char *description, void (*func) (void))
{
  double t1, t2, t;
  int i;

  //printf("Testing %s\n", description);
  t1 = get_time();
  for (i = 0; i < REPS; i++)
    func();
  XSync(disp, False);
  t2 = get_time();
  t = t2 - t1;
  printf("Time: %3.3f sec for %s.\n", t,description);
}

static Xrender_Surf *xrender_surf_new(Display * disp, Drawable draw, Visual * vis,
			       int w, int h, int alpha)
{
  Xrender_Surf *rs;
  XRenderPictFormat *fmt;
  XRenderPictureAttributes att;

  rs = calloc(1, sizeof(Xrender_Surf));
  if (alpha)
    fmt = XRenderFindStandardFormat(disp, PictStandardARGB32);
  else
    fmt = XRenderFindStandardFormat(disp, PictStandardRGB24);
  rs->w = w;
  rs->h = h;
  rs->depth = fmt->depth;
  rs->vis = vis;
  rs->draw = XCreatePixmap(disp, draw, w, h, fmt->depth);
  att.dither = 1;
  att.component_alpha = 1;
  att.repeat = 0;
  rs->pic =
      XRenderCreatePicture(disp, rs->draw, fmt,
			   CPRepeat | CPDither | CPComponentAlpha, &att);
  rs->allocated = 1;
  return rs;
}

static Xrender_Surf *xrender_surf_adopt(Display * disp, Drawable draw, Visual * vis,
				 int w, int h)
{
  Xrender_Surf *rs;
  XRenderPictFormat *fmt;
  XRenderPictureAttributes att;

  rs = calloc(1, sizeof(Xrender_Surf));
  fmt = XRenderFindVisualFormat(disp, vis);
  rs->w = w;
  rs->h = h;
  rs->depth = fmt->depth;
  rs->vis = vis;
  rs->draw = draw;
  att.dither = 1;
  att.component_alpha = 1;
  att.repeat = 0;
  rs->pic =
      XRenderCreatePicture(disp, rs->draw, fmt,
			   CPRepeat | CPDither | CPComponentAlpha, &att);
  rs->allocated = 0;
  return rs;
}

#if 0
static void xrender_surf_free(Display * disp, Xrender_Surf * rs)
{
  if (rs->allocated)
    XFreePixmap(disp, rs->draw);
  XRenderFreePicture(disp, rs->pic);
  free(rs);
}
#endif

static void xrender_surf_populate(Display * disp, Xrender_Surf * rs, int w, int h,
			   int *img_data)
{
  GC gc;
  XGCValues gcv;
  XImage *xim;
  int x, y;

  /* yes this isn't optimal - i know.. i just want some data for now */
  gc = XCreateGC(disp, rs->draw, 0, &gcv);
  xim =
      XCreateImage(disp, rs->vis, rs->depth, ZPixmap, 0, NULL, w, h, 32, 0);
  xim->data = malloc(xim->bytes_per_line * xim->height);
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x++) {
      int pixel;
      int a, r, g, b;

      pixel = img_data[(y * w) + x];
      a = (pixel >> 24) & 0xff;
      r = (pixel >> 16) & 0xff;
      g = (pixel >> 8) & 0xff;
      b = (pixel) & 0xff;
      r = (r * (a + 1)) / 256;
      g = (g * (a + 1)) / 256;
      b = (b * (a + 1)) / 256;
      XPutPixel(xim, x, y, (a << 24) | (r << 16) | (g << 8) | b);
    }
  }
  XPutImage(disp, rs->draw, gc, xim, 0, 0, 0, 0, w, h);
  free(xim->data);
  xim->data = NULL;
  XDestroyImage(xim);
  XFreeGC(disp, gc);
}

static void xrender_surf_blend(Display * disp, Xrender_Surf * src,
			Xrender_Surf * dst, int x, int y, int w, int h,
			int smooth)
{
    //XFilters *flt;
  XTransform xf;

  xf.matrix[0][0] = (65536 * src->w) / w;
  xf.matrix[0][1] = 0;
  xf.matrix[0][2] = 0;
  xf.matrix[1][0] = 0;
  xf.matrix[1][1] = (65536 * src->h) / h;
  xf.matrix[1][2] = 0;
  xf.matrix[2][0] = 0;
  xf.matrix[2][1] = 0;
  xf.matrix[2][2] = 65536;
  if (smooth)
    XRenderSetPictureFilter(disp, src->pic, "bilinear", NULL, 0);
  else
    XRenderSetPictureFilter(disp, src->pic, "nearest", NULL, 0);
  XRenderSetPictureTransform(disp, src->pic, &xf);
  XRenderComposite(disp, PictOpOver, src->pic, None, dst->pic, 0, 0, 0, 0, x,
		   y, w, h);
}

static void populate_from_file(Display * disp, Xrender_Surf * rs,
			       const char *file)
{
  Imlib_Image im;
  DATA32 *pixels;
  int w, h;

  im = imlib_load_image(file);
  imlib_context_set_image(im);
  pixels = imlib_image_get_data_for_reading_only();
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  xrender_surf_populate(disp, rs, w, h, pixels);
  imlib_image_put_back_data(pixels);
  imlib_free_image_and_decache();
}

static void test_over_x(void)
{
  int x, y;

  x = rand() % (xrender.win->w - xrender.img->w);
  y = rand() % (xrender.win->h - xrender.img->h);
  xrender_surf_blend(disp, xrender.img, xrender.win, x, y, xrender.img->w,
		     xrender.img->h, 1);
}

static void test_over_off_x(void)
{
  int x, y;

  x = rand() % (xrender.off->w - xrender.img->w);
  y = rand() % (xrender.off->h - xrender.img->h);
  xrender_surf_blend(disp, xrender.img, xrender.off, x, y, xrender.img->w,
		     xrender.img->h, 1);
}

static void test_over_imlib2(void)
{
  int x, y;
  int w, h, ww, hh;

  imlib_context_set_anti_alias(1);
  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  x = rand() % (ww - w);
  y = rand() % (hh - h);
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, x, y, w, h);
}

static void test_over_sdl(void)
{
  int x, y;
  SDL_Rect drect;

  x = rand() % (sdl.win->w - sdl.img->w);
  y = rand() % (sdl.win->h - sdl.img->h);
  drect.x=x;
  drect.y=y;

  SDL_BlitSurface(sdl.img, NULL, sdl.win, &drect);
}

static void test_over_gtk(void)
{
  int x, y;
  int win_w=gdk_pixbuf_get_width (pixbuf.win);
  int win_h=gdk_pixbuf_get_height (pixbuf.win);
  int img_w=gdk_pixbuf_get_width (pixbuf.img);
  int img_h=gdk_pixbuf_get_height (pixbuf.img);

  x = rand() % (win_w - img_w);
  y = rand() % (win_h - img_h);

  gdk_pixbuf_composite(pixbuf.img, pixbuf.win, x, y, img_w, img_h, x, y, 1, 1,
		       GDK_INTERP_NEAREST, 255);
}

static void test_over_ftwl(void)
{
  int x, y;
  struct ct_size win_size,img_size;
  struct ct_point pos;

  be_sprite_get_size(&img_size,ftwl.img);
  be_osda_get_size(&win_size,ftwl.win);

  x = rand() % (win_size.width - img_size.width);
  y = rand() % (win_size.height - img_size.height);
  pos.x=x;
  pos.y=y;

  be_draw_sprite(ftwl.win,ftwl.img,NULL,&pos,NULL);
}


static void main_loop(const char *transparent_filename)
{
#if 0
  /* printf query filters */
  printf("Available XRENDER filters:\n");
  {
    int i;
    XFilters *flt;

    flt = XRenderQueryFilters(disp, win);
    for (i = 0; i < flt->nfilter; i++)
      printf("%s\n", flt->filter[i]);
  }
#endif

  printf("Setup...\n");

  /* setup */

  // XRender
  {
    xrender.win =
	xrender_surf_adopt(disp, win,
			   DefaultVisual(disp, DefaultScreen(disp)),
			   WIN_WIDTH, WIN_HEIGHT);
    xrender.off =
	xrender_surf_new(disp, win, DefaultVisual(disp, DefaultScreen(disp)),
			 WIN_WIDTH, WIN_HEIGHT, 0);
    xrender.img =
	xrender_surf_new(disp, win, DefaultVisual(disp, DefaultScreen(disp)),
			 100, 100, 1);
    populate_from_file(disp, xrender.win, "tst_opaque.png");
    populate_from_file(disp, xrender.off, "tst_opaque.png");
    populate_from_file(disp, xrender.img, transparent_filename);
  }

  // imlib
  {
    imlib.win = imlib_load_image("tst_opaque.png");
    imlib.img = imlib_load_image(transparent_filename);
    assert(imlib.win && imlib.img);
  }

  // SDL
  if (DO_SDL) {
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
      fprintf(stderr, "SDL konnte nicht initialisiert werden:  %s\n",
	      SDL_GetError());
      exit(1);
    }

    atexit(SDL_Quit);

    sdl.win = IMG_Load("tst_opaque.png");
    sdl.img = IMG_Load(transparent_filename);
    assert(sdl.win && sdl.img);
  }


  // pixbuf
  {
      pixbuf.win = gdk_pixbuf_new_from_file("tst_opaque.png", NULL);
      pixbuf.img = gdk_pixbuf_new_from_file(transparent_filename, NULL);
  }

  // FTWL
  {
    struct Sprite *win_sprite = be_load_gfxfile("tst_opaque.png");
    struct ct_size size;

    be_sprite_get_size(&size,win_sprite);

    ftwl.img = be_load_gfxfile(transparent_filename);
    ftwl.win = be_create_osda(size.width,size.height);
    be_draw_sprite(ftwl.win, win_sprite, &size, NULL, NULL);
  }

  printf("*** ROUND 1 ***\n");

  srand(7);
  time_test("FTWL off-screen", test_over_ftwl);
  {
      be_copy_osda_to_screen(ftwl.win);
  }

  if (DO_SDL) {
    srand(7);
    time_test("SDL off-screen", test_over_sdl);
    {
      SDL_Rect dstrect;
      dstrect.x = 0;
      dstrect.y = 0;

      SDL_BlitSurface(sdl.win, NULL, sdl.display, &dstrect);
      SDL_UpdateRect(sdl.display, 0, 0, 0, 0);
    }
  }

  srand(7);
  time_test("Imlib2 off-screen", test_over_imlib2);
  {
    imlib_context_set_image(imlib.win);
    imlib_context_set_display(disp);
    imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
    imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
    imlib_context_set_drawable(win);
    imlib_render_image_on_drawable(0, 0);
  }

  srand(7);
  time_test("pixbuf off-screen", test_over_gtk);
  {
      GC gc;
      XGCValues gcv;

      gc = XCreateGC(disp, win, 0, &gcv);
      gdk_pixbuf_xlib_render_to_drawable(pixbuf.win, win, gc, 0, 0, 0, 0,
					 WIN_WIDTH, WIN_HEIGHT,
					 XLIB_RGB_DITHER_NORMAL, 0, 0);
  }

  srand(7);
  time_test("Xrender on-screen", test_over_x);

  srand(7);
  time_test("Xrender off-screen", test_over_off_x);

  printf("finished\n");
  sleep(200);
  exit(0);

#ifdef EXTENDED_TESTS
  printf("*** ROUND 2 ***\n");
  srand(7);
  //time_test("Test Xrender doing 1/2 scaled Over blends", test_over_scale_half_x);
  srand(7);
  //time_test("Test Xrender (offscreen) doing 1/2 scaled Over blends", test_over_off_scale_half_x);
  srand(7);
  time_test("Test Imlib2 doing 1/2 scaled Over blends",
	    test_over_scale_half_imlib2);
  imlib_context_set_image(imlib.win);
  imlib_context_set_display(disp);
  imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
  imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
  imlib_context_set_drawable(win);
  imlib_render_image_on_drawable(0, 0);
  sleep(2);
  printf("*** ROUND 3 ***\n");
  srand(7);
  //time_test("Test Xrender doing 2* smooth scaled Over blends", test_over_scale_double_smooth_x);
  srand(7);
  //time_test("Test Xrender (offscreen) doing 2* smooth scaled Over blends", test_over_off_scale_double_smooth_x);
  srand(7);
  time_test("Test Imlib2 doing 2* smooth scaled Over blends",
	    test_over_scale_double_smooth_imlib2);
  imlib_context_set_image(imlib.win);
  imlib_context_set_display(disp);
  imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
  imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
  imlib_context_set_drawable(win);
  imlib_render_image_on_drawable(0, 0);
  sleep(2);
  printf("*** ROUND 4 ***\n");
  srand(7);
  //time_test("Test Xrender doing 2* nearest scaled Over blends", test_over_scale_double_nearest_x);
  srand(7);
  //time_test("Test Xrender (offscreen) doing 2* nearest scaled Over blends", test_over_off_scale_double_nearest_x);
  srand(7);
  time_test("Test Imlib2 doing 2* nearest scaled Over blends",
	    test_over_scale_double_nearest_imlib2);
  imlib_context_set_image(imlib.win);
  imlib_context_set_display(disp);
  imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
  imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
  imlib_context_set_drawable(win);
  imlib_render_image_on_drawable(0, 0);
  printf("*** ROUND 6 ***\n");
  count = 0;
  time_test("Test Xrender doing general nearest scaled Over blends",
	    test_over_scale_general_nearest_x);
  count = 0;
  time_test
      ("Test Xrender (offscreen) doing general nearest scaled Over blends",
       test_over_off_scale_general_nearest_x);
  count = 0;
  time_test("Test Imlib2 doing general nearest scaled Over blends",
	    test_over_scale_general_nearest_imlib2);
  imlib_context_set_image(imlib.win);
  imlib_context_set_display(disp);
  imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
  imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
  imlib_context_set_drawable(win);
  imlib_render_image_on_drawable(0, 0);
  sleep(2);
  printf("*** ROUND 7 ***\n");
  count = 0;
  time_test("Test Xrender doing general smooth scaled Over blends",
	    test_over_scale_general_smooth_x);
  count = 0;
  time_test
      ("Test Xrender (offscreen) doing general smooth scaled Over blends",
       test_over_off_scale_general_smooth_x);
  count = 0;
  time_test("Test Imlib2 doing general smooth scaled Over blends",
	    test_over_scale_general_smooth_imlib2);
  imlib_context_set_image(imlib.win);
  imlib_context_set_display(disp);
  imlib_context_set_visual(DefaultVisual(disp, DefaultScreen(disp)));
  imlib_context_set_colormap(DefaultColormap(disp, DefaultScreen(disp)));
  imlib_context_set_drawable(win);
  imlib_render_image_on_drawable(0, 0);
  sleep(2);
#endif

  XSync(disp, False);
}

static void setup_window(void)
{
  XSetWindowAttributes att;
  XClassHint *xch;

  att.background_pixmap = None;
  att.colormap = DefaultColormap(disp, DefaultScreen(disp));
  att.border_pixel = 0;
  att.event_mask =
      ButtonPressMask |
      ButtonReleaseMask |
      EnterWindowMask |
      LeaveWindowMask |
      PointerMotionMask |
      ExposureMask | StructureNotifyMask | KeyPressMask | KeyReleaseMask;
  win = XCreateWindow(disp,
		      RootWindow(disp, DefaultScreen(disp)),
		      0, 0, WIN_WIDTH, WIN_HEIGHT, 0,
		      DefaultDepth(disp, DefaultScreen(disp)),
		      InputOutput,
		      DefaultVisual(disp, DefaultScreen(disp)),
		      CWColormap | CWBorderPixel | CWEventMask |
		      CWBackPixmap, &att);
  XStoreName(disp, win, "Xrender/imlib/pixbuf output");
  xch = XAllocClassHint();
  xch->res_name = "Main";
  xch->res_class = "Render_Demo";
  XSetClassHint(disp, win, xch);
  XFree(xch);
  XMapWindow(disp, win);
  XSync(disp, False);
  usleep(200000);
  XSync(disp, False);
}

#ifdef USE_GTK
/* Destroy handler for the window */
static void destroy_cb(GtkObject * object, gpointer data)
{
  gtk_main_quit();
}
#endif

int main(int argc, char **argv)
{
    if(argc!=2) {
	printf("Usage: render_bench <transparent_file.png>\n");
	return 0;
    }
  disp = XOpenDisplay(NULL);
  if (!disp) {
    printf("ERROR: Cannot connect to display!\n");
    exit(-1);
  }

  // SDL
  sdl.display = SDL_SetVideoMode(WIN_WIDTH, WIN_HEIGHT, 32, SDL_SWSURFACE);
  if (sdl.display == NULL) {
    fprintf(stderr, "Konnte kein Fenster 640x480 px oeffnen: %s\n",
	    SDL_GetError());
    exit(1);
  }
  SDL_WM_SetCaption("SDL output", NULL);

  // Pixbuf
  gdk_pixbuf_xlib_init(disp, DefaultScreen(disp));

  // FTWL
  {
    struct ct_size screen_size = { WIN_WIDTH, WIN_HEIGHT };

    be_init(&screen_size, FALSE);
  }
  
  setup_window();
  main_loop(argv[1]);
  return 0;
}

#ifdef EXTENDED_TESTS
static void test_over_scale_half_x(void)
{
  int x, y;

  x = rand() % (xrender.win->w - (xrender.img->w / 2));
  y = rand() % (xrender.win->h - (xrender.img->h / 2));
  xrender_surf_blend(disp, xrender.img, xrender.win, x, y, xrender.img->w / 2,
		     xrender.img->h / 2, 0);
}

static void test_over_off_scale_half_x(void)
{
  int x, y;

  x = rand() % (xrender.off->w - (xrender.img->w / 2));
  y = rand() % (xrender.off->h - (xrender.img->h / 2));
  xrender_surf_blend(disp, xrender.img, xrender.off, x, y, xrender.img->w / 2,
		     xrender.img->h / 2, 0);
}

static void test_over_scale_half_imlib2(void)
{
  int x, y;
  int w, h, ww, hh;

  imlib_context_set_anti_alias(0);
  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  x = rand() % (ww - (w / 2));
  y = rand() % (hh - (h / 2));
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, x, y, w / 2, h / 2);
}

static void test_over_scale_double_smooth_x(void)
{
  int x, y;

  x = rand() % (xrender.win->w - (xrender.img->w * 2));
  y = rand() % (xrender.win->h - (xrender.img->h * 2));
  xrender_surf_blend(disp, xrender.img, xrender.win, x, y, xrender.img->w * 2,
		     xrender.img->h * 2, 1);
}

static void test_over_off_scale_double_smooth_x(void)
{
  int x, y;

  x = rand() % (xrender.off->w - (xrender.img->w * 2));
  y = rand() % (xrender.off->h - (xrender.img->h * 2));
  xrender_surf_blend(disp, xrender.img, xrender.off, x, y, xrender.img->w * 2,
		     xrender.img->h * 2, 1);
}

static void test_over_scale_double_smooth_imlib2(void)
{
  int x, y;
  int w, h, ww, hh;

  imlib_context_set_anti_alias(1);
  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  x = rand() % (ww - (w * 2));
  y = rand() % (hh - (h * 2));
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, x, y, w * 2, h * 2);
}

static void test_over_scale_double_nearest_x(void)
{
  int x, y;

  x = rand() % (xrender.win->w - (xrender.img->w * 2));
  y = rand() % (xrender.win->h - (xrender.img->h * 2));
  xrender_surf_blend(disp, xrender.img, xrender.win, x, y, xrender.img->w * 2,
		     xrender.img->h * 2, 0);
}

static void test_over_off_scale_double_nearest_x(void)
{
  int x, y;

  x = rand() % (xrender.off->w - (xrender.img->w * 2));
  y = rand() % (xrender.off->h - (xrender.img->h * 2));
  xrender_surf_blend(disp, xrender.img, xrender.off, x, y, xrender.img->w * 2,
		     xrender.img->h * 2, 0);
}

static void test_over_scale_double_nearest_imlib2(void)
{
  int x, y;
  int w, h, ww, hh;

  imlib_context_set_anti_alias(0);
  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  x = rand() % (ww - (w * 2));
  y = rand() % (hh - (h * 2));
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, x, y, w * 2, h * 2);
}

static int count = 0;

static void test_over_scale_general_nearest_x(void)
{
  int w, h;

  w = 1 + ((xrender.img->w * count) / (REPS / 16));
  h = 1 + ((xrender.img->h * count) / (REPS / 16));
  xrender_surf_blend(disp, xrender.img, xrender.win, 0, 0, w, h, 0);
  count++;
}

static void test_over_off_scale_general_nearest_x(void)
{
  int w, h;

  w = 1 + ((xrender.img->w * count) / (REPS / 16));
  h = 1 + ((xrender.img->h * count) / (REPS / 16));
  xrender_surf_blend(disp, xrender.img, xrender.off, 0, 0, w, h, 0);
  count++;
}

static void test_over_scale_general_nearest_imlib2(void)
{
  int w, h, ww, hh;

  imlib_context_set_anti_alias(0);
  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  ww = 1 + ((w * count) / (REPS / 16));
  hh = 1 + ((h * count) / (REPS / 16));
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, 0, 0, ww, hh);
  count++;
}

static void test_over_scale_general_smooth_x(void)
{
  int w, h;

  w = 1 + ((xrender.img->w * count) / (REPS / 16));
  h = 1 + ((xrender.img->h * count) / (REPS / 16));
  xrender_surf_blend(disp, xrender.img, xrender.win, 0, 0, w, h, 1);
  count++;
}

static void test_over_off_scale_general_smooth_x(void)
{
  int w, h;

  w = 1 + ((xrender.img->w * count) / (REPS / 16));
  h = 1 + ((xrender.img->h * count) / (REPS / 16));
  xrender_surf_blend(disp, xrender.img, xrender.off, 0, 0, w, h, 1);
  count++;
}

static void test_over_scale_general_smooth_imlib2(void)
{
  int w, h, ww, hh;

  imlib_context_set_image(imlib.win);
  ww = imlib_image_get_width();
  hh = imlib_image_get_height();
  imlib_context_set_image(imlib.img);
  w = imlib_image_get_width();
  h = imlib_image_get_height();
  ww = 1 + ((w * count) / (REPS / 16));
  hh = 1 + ((h * count) / (REPS / 16));
  if ((ww < w) && (hh < h))
    imlib_context_set_anti_alias(0);
  else
    imlib_context_set_anti_alias(1);
  imlib_context_set_image(imlib.win);
  imlib_blend_image_onto_image(imlib.img, 0, 0, 0, w, h, 0, 0, ww, hh);
  count++;
}
#endif
